(*                         J is for JavaScript:
 *                    A direct-style correspondence
 *             between Algol-like languages and JavaScript
 *                   using first-class continuations
 *
 *            Olivier Danvy, Chung-chieh Shan, and Ian Zerny
 *                            Copyright 2009
 *                  GPLv3, see LICENSE.txt for details
 *
 *                    www.zerny.dk/j-is-for-js.html
 *
 *                         Version: April 2011
 *)

(* Constructor definitions for our "Pidgin" Algol. *)
(* Each construct is homomorphically mapped to the corresponding code *)
(* in JavaScript. *)

type var = Var of string
type lbl = Lbl of string
type def = Def of string
type cmd = Cmd of string
type exp = Exp of string
type prg = Prg of string

let unVar (Var x) = x
let unLbl (Lbl x) = x
let unDef (Def x) = x
let unCmd (Cmd x) = x
let unExp (Exp x) = x
let unPrg (Prg x) = x

let defVar : var -> def
  = fun (Var v) -> Def ("var " ^ v ^ ";\n")

let defFun : var -> var list -> def list -> cmd list -> def
  = fun var args defs cmds -> Def
    ( "var "
      ^ unVar var
      ^ " = function ("
      ^ String.concat ", " (List.map unVar args)
      ^ ") {\nvar JI = new Continuation();\n"
      ^ String.concat "" (List.map unDef defs)
      ^ "return function () {\n"
      ^ String.concat "" (List.map unCmd cmds)
      ^ "};\n};\n" )

let defLbl : lbl -> cmd list -> def
  = fun lbl cmds -> Def
    ( "var "
      ^ unLbl lbl
      ^ " = function () { JI(function () {\n"
      ^ String.concat "" (List.map unCmd cmds)
      ^ "});};\n" )

let cmdBlock : def list -> cmd list -> cmd
  = fun defs cmds -> Cmd
    ( "{\n"
      ^ String.concat "" (List.map unDef defs)
      ^ String.concat "" (List.map unCmd cmds)
      ^ "}\n" )

let cmdAssign : exp -> exp -> cmd
  = fun (Exp l) (Exp r) -> Cmd (l ^ " = " ^ r ^ ";\n")

let cmdGoto : lbl -> cmd
  = fun (Lbl l) -> Cmd (l ^ "();\n")

let cmdYield : lbl -> lbl -> cmd
  = fun (Lbl src) (Lbl dst) -> Cmd
    (* for coroutine use, src and dst must be bound at the same lexical scope *)
    ( "(function () { "
    ^ src
    ^ " = new Continuation(); "
    ^ dst
    ^ "(); })();\n" )

let cmdReturn : exp -> cmd
= fun (Exp e) -> Cmd ("return " ^ e ^ ";\n")

let cmdIfThen : exp -> cmd -> cmd
= fun (Exp e) (Cmd c) -> Cmd ("if (" ^ e ^ ") " ^ c)

let cmdExp : exp -> cmd
= fun (Exp e) -> Cmd (e ^ ";\n")

let cmdPrint : exp -> cmd
= fun (Exp e) -> Cmd ("print(" ^ e ^ ");\n")

let expVar : var -> exp
= fun (Var v) -> Exp v

let expChr : char -> exp
= fun c -> Exp ("'"^ String.make 1 c ^"'")

let expStr : string -> exp
= fun s -> Exp ("\"" ^ s ^ "\"")

let expStrConcat : exp -> exp -> exp
= fun (Exp s1) (Exp s2) -> Exp (s1 ^ " + " ^ s2)

let expCall : var -> exp list -> exp
= fun var exps -> Exp
    ( unVar var
    ^ "("
    ^ String.concat ", " (List.map unExp exps)
    ^ ")()" )

let expLength : exp -> exp
= fun (Exp e) -> Exp (e ^ ".length")

let expInt : int -> exp
= fun i -> Exp (string_of_int i)

let expBool : bool -> exp
= function true -> Exp "true" | false -> Exp "false"

let expArray : exp list -> exp
= fun inits -> Exp ("[" ^ String.concat ", " (List.map unExp inits) ^ "]")

let expBottom : exp
= Exp "undefined"

let expBinary : string -> exp -> exp -> exp
= fun op (Exp e1) (Exp e2) -> Exp (e1 ^ " " ^ op ^ " " ^ e2)

let expEq = expBinary "=="
let expNotEq = expBinary "!="
let expLt = expBinary "<"
let expGt = expBinary ">"
let expAdd = expBinary "+"
let expSub = expBinary "-"
let expMul = expBinary "*"

let expIndex : exp -> exp -> exp
= fun (Exp e1) (Exp e2) -> Exp (e1 ^ "[" ^ e2 ^ "]")

let program : def list -> cmd list -> prg
= fun defs cmds ->
  let Def d = defFun (Var "main") [] defs cmds in
  let Cmd c = cmdExp (expCall (Var "main") []) in
    Prg ( d ^ c )

let print_program : prg -> unit
= fun (Prg p) -> print_string p
